//
//  Lambda:
//  . CApplication
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//

#include "application.h"
#include "macro.h"

int unify_count = 0;

int i_reduce, k_reduce, s_reduce;

/***** CApplication *****/

//
//  reduziert "((lambda x.t) a)" zu "t[x/a]" (falls mglich)
//
CNode* CApplication::reduce0(bool red_expr)
{
    if(!expr || !value)
        return 0;

    CLambda* pl = 0;
    if(reduction_order == EAGER_REDUCTION) {
        CNode* left = (!red_expr) ? move(expr) : do_reduce(expr);
        CNode* right = (changed || !red_expr) ? move(value) : do_reduce(value);
        if(changed || !(pl = left->lambda()))
            return new CApplication(left, right);
        delete expr;
        delete value;
        expr = left;
        value = right;
    } else {
        pl = expr->lambda();
        if(!pl) {
            CNode* left = (!red_expr) ? move(expr) : do_reduce(expr);
            CNode* right = (!red_expr || (reduction_order != BROAD_REDUCTION && changed)) ? move(value) : do_reduce(value);
            return new CApplication(left, right);
        }
    }
    // pl ist nun (lambda x.t)
    if(!pre_alpha_flag) {
        TVariable *fv = value->freevars(0);
        if(fv) 
            expr->alpha(fv);
        deletelist(fv);
    }
    
    changed = true;
    n_beta++;
    return pl->transform(value);
}

CNode* CApplication::reduce()
{
    return reduce0(true);
}

CNode* CApplication::eta(int stage, std::string s)
{
    if(stage != 0)
        return 0;
    if(!value->eta(1, s))
        return 0;
    
    TVariable* fv = expr->freevars(0);
    for(TVariable* i = fv; i != 0; i = i->next)
        if(i->name == s)
            return 0;
    deletelist(fv);
    
    return move(expr);
}

CNode* CApplication::copy()
{
    return new CApplication(expr->copy(), value->copy());
}

void CApplication::print(std::ostream& out, TPrintContext pc)
{
    if(full_paren_flag) {
	out << "(";
	expr->print(out);
	out << " ";
	value->print(out);
	out << ")";
    } else {
	if(pc==PRINT_LEFT_APP || pc==PRINT_ABSTRACT) {
	    expr->print(out,PRINT_LEFT_APP);
	    out << " ";
	    value->print(out,PRINT_RIGHT_APP);
	} else {
	    out << "(";
	    expr->print(out,PRINT_LEFT_APP);
	    out << " ";
	    value->print(out,PRINT_RIGHT_APP);
	    out << ")";
	}
    }
}

TVariable* CApplication::freevars(TVariable *p)
{
    p=expr->freevars(p);
    p=value->freevars(p);
    return p;
}

bool CApplication::alpha(TVariable *fv)
{
    WHERE;
    bool res;
    res = expr->alpha(fv);
    if(value->alpha(fv)) res=true;
    return res;
}

void CApplication::global_alpha(TVariable* fv, int action)
{
    if(expr)
        expr->global_alpha(fv, action);
    if(value)
        value->global_alpha(fv, action);
}

CNode* CApplication::replace(std::string& sym, CNode* rep)
{
    return new CApplication(do_replace(expr,sym,rep), do_replace(value,sym,rep));
}

bool CApplication::numcheck(TNumCheck& nc)
{
    if(nc.status==TNumCheck::CHECK_RIGHT) {
	nc.status=TNumCheck::CHECK_LEFT;
	if(!expr->numcheck(nc)) return false;
	nc.status=TNumCheck::CHECK_RIGHT;
	nc.value++;
	return value->numcheck(nc);
    }
    return false;
}

bool CApplication::listcheck(TListCheck& lc)
{
    if(lc.status == TListCheck::CHECK_APP) {
	lc.status = TListCheck::CHECK_1ST_APP;
	if(!expr->listcheck(lc)) return false;
	lc.status = TListCheck::CHECK_TAIL;
	return value->listcheck(lc);
    } else if(lc.status == TListCheck::CHECK_1ST_APP) {
	lc.status = TListCheck::CHECK_1ST;
	if(!expr->listcheck(lc)) return false;
	lc.status = TListCheck::CHECK_2ND;
	return value->listcheck(lc);
    } else
	return lc.status == TListCheck::CHECK_2ND;
}

void CApplication::listout(std::ostream& o, int n, char delim)
{
    if(n == TListCheck::CHECK_APP) {
	expr->listout(o,TListCheck::CHECK_1ST_APP,delim);
	value->listout(o,TListCheck::CHECK_TAIL,' ');
    }
    else if(n == TListCheck::CHECK_1ST_APP) {
	struct TNumCheck nc;
	nc.status = TNumCheck::CHECK_X;
	nc.value = 0;
	o << delim;
	if(value->numcheck(nc) && (nc.value>=0 || true_ok)) {
	    if(nc.value==-1) o << "<true>";
	    else o << nc.value;
	} else {
	    if(!try_print_list(o,value)) value->print(o);
	}
    }
}

CTypePtr CApplication::get_type(TTypeInfo& ti, TTypeItem* local_types)
{
    /*
     *  type(M N) = 
     *    wenn type(M) = a -> 
     *    und  type(N) = a
     */
    CTypePtr m = expr->get_type(ti, local_types);
    if(!m)
        return 0;
    CTypePtr n = value->get_type(ti, local_types);
    if(!n)
        return 0;
    m.sync();           // damit sind m und n keine Aliase auf andere Typen
    n.sync();

    switch(m->get_type()) {
     case CType::TYPE_ATOM:
        /* m ist ein Atom -> Typfehler */
        ti.error_msg = "Typkonflikt: `" + m->get_name() + "' als Funktion verwendet.";
        return 0;
     case CType::TYPE_UNKNOWN:
        /* m hat keinen bekannten Typ -> konstruiere einen */
        {
            if(n->contains(m->get_ptr())) {
                ti.error_msg = "Rekursiver Typ [3]";
                return 0;
            }
            CTypePtr beta = new CType();
            m->make_function(n, beta);
            return beta;
        }
     case CType::TYPE_FUNCTION:
        /* m ist ein Funktionstyp -- Typen mssen passen */
        unify_count++;
        if(!m->get_in()->unify(n, ti.error_msg))
            return 0;
        return m->get_out();
        
     case CType::TYPE_ALIAS:
        cerr << "can't happen" << endl;
        exit(1);
    }
    /* not reached */
    return 0;
}

CNode* CApplication::ski()
{
    CNode* a = expr->ski();
    CNode* b = value->ski();
    if(a && b)
        return new CApplication(a, b);
    else
        return 0;
}

CNode* CApplication::ski_lambda(std::string s)
{
    /* Abkrzung: `?x.<foobar>' == `K <foobar>' wenn
       <foobar> unabhngig von x; standardmig wird
       `?x.a b' zu `S (K a) (K b)', mit Abkrzung
       ergibt sich direkt `K (a b)' */
    if(ski_shortcut_flag) {
        TVariable *fv = freevars(0);
        TVariable *p = fv;
        while(p && p->name != s)
            p = p->next;
        deletelist(fv);
        if(!p) {
            CNode* new_x = ski();
            if(new_x)
                return new CApplication(new CMacro("K"), new_x);
            else
                return 0;
        }
    }
    
    CNode* new_a = expr->ski();
    CNode* new_b = value->ski();

    CNode* res = 0;
    if(new_a && new_b) {
        CNode* a = new_a->ski_lambda(s);
        CNode* b = new_b->ski_lambda(s);
        if(a && b)
            res = new CApplication(new CApplication(new CMacro("S"), a), b);
    }
    delete new_a;
    delete new_b;
    return res;
}

CNode* CApplication::ski_reduce()
{
    if(reduction_order == EAGER_REDUCTION) {
        CNode* new_a = expr->ski_reduce();
        CNode* new_b = changed ? move(value) : value->ski_reduce();
        if(changed)
            return new CApplication(new_a, new_b);
        delete expr;
        delete value;
        expr = new_a;
        value = new_b;
    }
    if(expr->is_ski('I')) {
        changed = true;
        i_reduce++;
        return move(value);
    };
    CApplication* a = expr->applic();
    if(a) {
        if(a->expr->is_ski('K')) {
            changed = true;
            k_reduce++;
            return move(a->value);
        }
        CApplication* b = a->expr->applic();
        if(b && b->expr->is_ski('S')) {
            changed = true;
            s_reduce++;
            CNode* c = value->copy();
            return new CApplication(new CApplication(move(b->value), c),
                                    new CApplication(move(a->value), move(value)));
        }
    }
    if(reduction_order == EAGER_REDUCTION)
        return new CApplication(move(expr), move(value));
    
    /* keine Reduktion hier -- Unterobjekte reduzieren */
    CNode* new_a = expr->ski_reduce();
    CNode* new_b = (changed && reduction_order != BROAD_REDUCTION) ? move(value) : value->ski_reduce();
    return new CApplication(new_a, new_b);
}

CNode* CApplication::ioreduce(std::ostream& os, TIoContext& ioc)
{
    if(!expr || !value)
        return 0;

    CNode* pl = expr->lambda();
    if(!pl) {
        // links ist kein Lambda-Ausdruck
        //expr = expr->reduce();
        expr = do_reduce(expr);
        if(!changed) {
            if(full_paren_flag || (ioc.pc != PRINT_LEFT_APP && ioc.pc != PRINT_ABSTRACT)) {
                os << "(";
                ioc.parens_needed++;
            } 
            expr->print(os, PRINT_LEFT_APP);
            os << " " << flush;
            ioc.pc = PRINT_RIGHT_APP;
            return value->ioreduce(os, ioc);
        }
    }
    return reduce0(false);
}

bool CApplication::calls_macro(std::string s)
{
    return expr->calls_macro(s) || value->calls_macro(s);
}

CNode* CApplication::replace_macro(std::string mac, std::string sym)
{
    expr  = expr->replace_macro(mac, sym);
    value = value->replace_macro(mac, sym);
    return this;
}
