//
//  Lambda:
//  . CMacro - Makroersatz
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//
#include <iostream>
#include <string>
#include <map>
#include "macro.h"
#include "main.h"
#include "application.h"
#include "parse.h"
#include "scanner.h"
#include "symbol.h"

macros_t macros;	// Symboltabelle

/***** CMacro *****/
CNode* CMacro::reduce()
{
    if(!value) {
        makevalue();
        if(value)
            return move(value);
        else
            return copy();
    } else
        return do_reduce(value);
}

CNode* CMacro::copy()
{
    if(value)
        return value->copy();
    else
        return new CMacro(name, 0);
}

void CMacro::print(std::ostream& out, TPrintContext pc)
{
    if(value)
        value->print(out, pc);
    else if(magic_combinators && name.length() == 1)
        out << name;
    else
        out << "<" << name << ">";
}

CLambda* CMacro::lambda()
{
    if(!value) return 0;
    return value->lambda();
}

CNode* CMacro::eta(int stage, std::string s)
{
    if(!value)
        return 0;
    return value->eta(stage, s);
}

TVariable* CMacro::freevars(TVariable* p)
{
    if(!value) return p;
    return value->freevars(p);
}

bool CMacro::alpha(TVariable* fv)
{
    if(value) return value->alpha(fv);
    return false;
}

void CMacro::global_alpha(TVariable* fv, int action)
{
    if(value)
        value->global_alpha(fv, action);
}

void CMacro::makevalue()
{
    macros_t::const_iterator t = macros.find(name);
    if(t != macros.end()) {
	value = t->second.contents->copy();
        if(pre_alpha_flag)
            value->global_alpha(0, -1);
        n_expand++;
	changed = true;
    } else {
        std::cerr << "Fehler: Undefiniertes Makro <" << name << ">." << std::endl;
        abort_reduction = true;
    }
}

CNode* CMacro::replace(std::string& sym, CNode* rep)
{
    if(value)
        return do_replace(value,sym,rep);
    else
        return copy();
}

bool CMacro::numcheck(TNumCheck& nc)
{
    if(value) return value->numcheck(nc);
    else return false;
}

CTypePtr CMacro::get_type(TTypeInfo& ti, TTypeItem* local_types)
{
    if(ctrlcpressed) {
        ti.error_msg = "Makro-Expansion abgebrochen in <" + name + ">.";
        return 0;
    }
    if(!value)
        makevalue();
    if(!value)
        return 0;       
    return value->get_type(ti, local_types);
}

CNode* CMacro::ski()
{
    if(ctrlcpressed) {
        ctrlcpressed = false;
        error_message() << "Makro-Expansion abgebrochen in <" << name << ">.\n";
        return 0;
    }
    if(!magic_id(name) && !value)
        makevalue();
    if(value)
        return value->ski();
    else
        return new CMacro(name);
}

CNode* CMacro::ski_lambda(std::string s)
{
    if(ctrlcpressed) {
        ctrlcpressed = false;
        error_message() << "Makro-Expansion abgebrochen in <" << name << ">.\n";
        return 0;
    }
    if(!magic_id(name) && !value)
        makevalue();
    if(value)
        return value->ski_lambda(s);
    else
        return new CApplication(new CMacro("K"), new CMacro(name));
}

CNode* CMacro::ski_reduce()
{
    return new CMacro(name);
}

bool CMacro::calls_macro(std::string s)
{
    if(value)
        return value->calls_macro(s);
    if(name == s)
        return true;
    
    /* wird dieses Makro rekursiv aufgerufen? */
    macros_t::iterator i = macros.find(name);
    if(i == macros.end())
        return false;           // Makro nicht definiert
    if(i->second.flags & MAC_RECURSE)
        return false;           // Rekursion erkannt, aber nicht
                                // ber dieses Makro

    CProtector p(i->second.flags, MAC_RECURSE);
    return i->second.contents->calls_macro(s);
}

CNode* CMacro::replace_macro(std::string mac, std::string sym)
{
    if(value) {
        /* should not happen */
        value = value->replace_macro(mac, sym);
        return this;
    }
    if(name == mac) {
        /* Makrorekursion */
        delete this;
        return new CSymbol(sym);
    }

    macros_t::iterator i = macros.find(name);
    if(i == macros.end())
        return this;
    if(i->second.contents->calls_macro(mac)) {
        /* Indirekte Makrorekursion */
        CNode* a = i->second.contents->copy();
        delete this;
        return a->replace_macro(mac, sym);
    }
    return this;
}
