//
//  Lambda:
//  . CMacro - Makroersatz
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//
#ifndef LAMBDA_H
#define LAMBDA_H

#include <string>
#include <iostream>
#include <map>
#include "node.h"

// Makro "<name>"
// oder bereits ersetzt durch value
class CMacro : public CNode {
    std::string name;
    CNode* value;
 public:
    CMacro(const std::string& aname, CNode *aval=0) : name(aname), value(aval) {}
    virtual ~CMacro() { if(value) delete value; }

    virtual CNode* eta(int stage, std::string s);
    virtual CNode* reduce();
    virtual CLambda* lambda();
    virtual CNode* copy();
    virtual void print(std::ostream& out, TPrintContext pc);
    virtual TVariable* freevars(TVariable* p);
    virtual bool alpha(TVariable* fv);
    virtual void global_alpha(TVariable* fv, int action);
    virtual CNode* replace(std::string& sym, CNode* rep);
    virtual bool numcheck(TNumCheck& nc);
    virtual bool listcheck(TListCheck& lc) { return false; }
    virtual int count_atoms() { return value?value->count_atoms():1; }
    virtual CTypePtr get_type(TTypeInfo& ti, TTypeItem* local_types);
    virtual CNode* ski();
    virtual CNode* ski_lambda(std::string);
    void makevalue();

    bool is_ski(char c) { return name.length() == 1 && name[0] == c; }
    CNode* ski_reduce();

    bool calls_macro(std::string);
    virtual CNode* replace_macro(std::string mac, std::string sym);
};

struct TMacro {
    CNode* contents;
    std::string doc;
    int flags;
    TMacro(CNode* acont=0, std::string adoc=std::string())
        : contents(acont), doc(adoc), flags(0) {}
};

enum {
    /* Bitflags fr TMacro::flags */
    MAC_RECURSE = 1
};

typedef std::map<std::string,TMacro> macros_t;
extern macros_t macros;

#endif
