//
//  Lambda - Hauptprogramm
//
//  (c) copyright 1998,1999,2000 by Stefan Reuther <Streu@gmx.de>
//
#define _POSIX_SOURCE

#include <iostream>
#include <string>
#include <fstream>
#include <signal.h>
#include <time.h>
#include <cstdlib>
#include <algorithm>
#include "scanner.h"
#include "macro.h"
#include "application.h"
#include "symbol.h"
#include "node.h"
#include "number.h"
#include "rl-stuff.h"
#include "main.h"
#include "parse.h"

using namespace std;

#define countof(x) (int)(sizeof(x)/sizeof(x[0]))

#define VERSION "2.8"
#ifndef DIRECTORY
# define DIRECTORY
#endif
#define STRINGIFY_X(x) #x
#define STRINGIFY(x) STRINGIFY_X(x)

#define INV_ON_CS "\033[7m"
#define INV_OFF_CS "\033[0m"
#ifdef WITH_READLINE
# define INV_ON "\001\033[7m\002"
# define INV_OFF "\001\033[0m\002"
# define RL_COMPLETE(x) ,x
#else
# define INV_ON INV_ON_CS
# define INV_OFF INV_OFF_CS
# define RL_COMPLETE(x) /**/
#endif

bool verbose_flag = false;    // jeden Schritt ausgeben?
bool wait_flag    = false;    // nach jedem Schritt auf [CR] warten?
bool number_flag  = true;     // Zahlenreprsentation auflsen im Endergebnis
bool timer_flag   = false;    // Auswertungszeit stoppen?
bool alarm_flag   = false;    // Beep am Ende?
bool inverse_flag = true;     // Prompt invers
bool eta_flag     = true;     // Eta-Reduktion?
bool ctrlc_mark_flag = false; // Abbruchstelle markieren?

CNode* result = 0;

volatile bool ctrlcpressed = false;	// Abbruch mit C-c ist mglich
volatile bool reading_stdin = false;

bool abort_reduction = false;

int default_command = 0;        // 0 = eval

#ifdef WITH_SIGACTION
struct sigaction set,get;
#endif

/* wird bei SIGINT aka C-c aufgerufen */
void ctrlchandler(int)
{
    ctrlcpressed = true;
#ifdef WITH_SIGACTION
    sigaction(SIGINT,&set,&get);
#else
    signal(SIGINT, ctrlchandler);
#endif
    if(reading_stdin) {
	cout << " [`exit' eingeben fr Ende]" << endl << prompt;
	cout.flush();
    }
}

/* Resultat "reverse-engineeren" */
void pretty_print_result(CNode* p)
{
    if(number_flag) {
        struct TNumCheck nc;
        nc.status = TNumCheck::CHECK_X;
        nc.value = 0;
        if(p->numcheck(nc)) {
            if(nc.value == -1)
                cout << "Reprsentation des Wahrheitswertes `true'" << endl;
            else {		
                cout << "Reprsentation der Zahl " << nc.value << endl;
                if(nc.value==0)
                    cout << "Reprsentation des Wahrheitswertes `false'" << endl;
            }
        }
        
        if(try_print_list(cout,p,true)) cout << endl;
    }
}

// Nachbearbeitung eines Auswertungsschrittes
void after_eval()
{
    if(wait_flag && changed) {
        int c;
        while((c=cin.get())!=EOF && c!='\n' && cin) ;
        if(!cin || c==EOF) {
            cin.clear();       // Fehler ignorieren -> erlaubt Abbruch
            cout.clear();
            ctrlcpressed=true; // mit C-c oder EOF
            cout << endl;
        }
    } else {
        cout << endl;
    }
}

//
//  Ausdruck in p reduzieren
//
void evaluate(CNode* p)
{
    int step=0;
    result = 0;
    time_t starttime = time(0);
    time_t endtime = starttime;

    int o_beta = n_beta, o_eta = n_eta, o_expn = n_expand;

    // reduzieren
    ctrlcpressed = false;
    abort_reduction=false;
    if(pre_alpha_flag)
        p->global_alpha(0, -1);

    while(!ctrlcpressed && !abort_reduction) {
	changed=false;
	CNode* q;
        int k_beta = n_beta, k_eta = n_eta, k_expn = n_expand;
	q=CNode::do_reduce(p);
        cout.clear();
	step++;
	if(verbose_flag || wait_flag || !changed) {
	    if(changed) {
                cout << step;
                char prefix = '[';
                if(k_beta != n_beta)
                    cout << prefix << n_beta - k_beta << "",
                    prefix = ',';
                if(k_eta != n_eta)
                    cout << prefix << n_eta - k_eta << "n",
                    prefix = ',';
                if(k_expn != n_expand)
                    cout << prefix << n_expand - k_expn << "e",
                    prefix = ',';
                if(prefix != '[')
                    cout << "]: ";
                else
                    cout << ": ";
            }
            else if(post_alpha_flag)
                q->global_alpha(0, 0);
	    q->print(cout);
            after_eval();
	}
	p=q;
	if(!changed) break;
    }

    // Auswertung
    if(timer_flag) {
	endtime = time(0);
	cout << "Gesamtzeit: " << difftime(endtime,starttime)
             << " sec, " << step << " Schritte ("
             << n_beta - o_beta << " + " << n_eta - o_eta << "n + "
             << n_expand - o_expn << "e = "
             << n_beta+n_eta+n_expand - o_beta-o_eta-o_expn << ")"
             << endl;
    }
    if(alarm_flag) {
	cerr << "\a" << flush;
    }
    if(ctrlcpressed && changed) {
	cerr << "Auswertung abgebrochen nach " << step
             << " Schritten." << endl;
    } else
        pretty_print_result(p);

    result = p;
}

//
//  Kommando "eval term" bzw. nur "term": Term reduzieren
//
bool eval_func()
{
    CNode* p = doreadlambda();
    if(!p) return false;
    delete result;
    evaluate(p);
    return true;
}

//
//  "cont"-Kommando: letztes Ergebnis weiter auswerten
//
bool cont_func()
{
    if(!result) {
	error_message() << "Kein Ergebnis gespeichert." << endl;
	return true;
    }
    evaluate(result);
    return true;
}

//
//  "result"-Funktion: letztes Ergebnis anzeigen
//
bool result_func()
{
    if(!result) {
	error_message() << "kein Ergebnis gespeichert." << endl;
    } else {
	result->print(cout);
	cout << endl;
        pretty_print_result(result);
    }
    return true;
}

/* Parser fr def-Kommando */
bool define_helper(const char* func_name,
                   std::string& doc_str,
                   std::string& mname,
                   CNode*& node)
{
    if(magic_combinators && ttype == TSYMBOL && magic_id(tstr))
        ;
    else if(ttype!=TMACRO) {
	error_message() << "Nach `" << func_name << "' mu ein Makroname stehen." << endl;
	readeol();
	return false;
    }
    
    string s = tstr;
    getsymbol();

    if(ttype==TCOLON) {
        bool freeform_save = freeform_flag;
        freeform_flag = true;
	if(getsymbol() != TSYMBOL) {
	    error_message() << "Nach `" << func_name << " <" << s << "> :' mu der Dokumentationstext folgen." << endl;
	    readeol();
	    return false;
	}
	while(ttype==TSYMBOL) {
	    doc_str += "  ";
	    doc_str += tstr;
	    doc_str += "\n";
	    getsymbol();
	}
	freeform_flag = freeform_save;
	if(ttype != TCOLON) {
	    error_message() << "Der Dokumentationstext mu mit `:' abgeschlossen werden." << endl;
	    readeol();
	    return false;
	}
	getsymbol();
    }
	
    CNode *q = doreadlambda();
    if(!q)
        return false;
    node = q;
    mname = s;
//     if(node->calls_macro(mname))
//         cout << "Makrorekursion: Selbstaufruf in " << mname << endl;
    return true;
}

/* Makro definieren */
void define_macro(string s, string doc_str, CNode* q)
{
    macros_t::iterator t = macros.find(s);
    if(t!=macros.end())
        delete t->second.contents;

    if(post_alpha_flag)
        q->global_alpha(0, 0);

    macros[s] = TMacro(q,doc_str);
    if(s=="true") {
	TNumCheck nc;
	nc.status = TNumCheck::CHECK_X;
	nc.value = 0;
	true_ok = (q->numcheck(nc) && nc.value==-1);
    }
    if(verbose_flag) cout << "Makro <" << s << "> definiert." << endl;
}

//
//  Kommando "def <macro> term" definiert ein Makro
//
bool define_func()
{
    string doc_str;
    string mac_name;
    CNode* q;

    if(!define_helper("def", doc_str, mac_name, q))
        return false;

    define_macro(mac_name, doc_str, q);
    return true;
}

//
//  Kommando "edef <macro> term" definiert ein Makro, aber der Term
//  wird vorher expandiert.
//
bool edef_func()
{
    string doc_str;
    string mac_name;
    CNode* p;

    if(!define_helper("edef", doc_str, mac_name, p))
        return false;

    ctrlcpressed = false;
    abort_reduction = false;
    if(pre_alpha_flag)
        p->global_alpha(0, -1);
    while(!ctrlcpressed && !abort_reduction) {
	changed=false;
	CNode* q;
	q = CNode::do_reduce(p);
	p = q;
	if(!changed) break;
    }

    cout.clear();
    if(ctrlcpressed) {
        cerr << "Expansion von <" << mac_name << "> abgebrochen" << endl;
        readeol();
        return false;
    }

    define_macro(mac_name, doc_str, p);
    return true;
}

//
//  Kommando "rdef <macro> term"
//
bool rdef_func()
{
    string doc_str;
    string mac_name;
    CNode* q;

    if(!define_helper("rdef", doc_str, mac_name, q))
        return false;

    if(q->calls_macro(mac_name)) {
        if(verbose_flag)
            cout << "Selbstaufruf in <" << mac_name << ">" << flush;

        string sym = newtemp();
        q = q->replace_macro(mac_name, sym);
        q = new CApplication(new CMacro("Y"),
                             new CLambda(sym, q, 0));
        
        if(verbose_flag)
            cout << endl;
    }

    define_macro(mac_name, doc_str, q);
    return true;
}

/* Datei einlesen und ausfhren */
bool do_read(string fname)
{
    bool append=true;
    for(int i=fname.length()-1; i>=0; i--) {
	if(fname[i]=='/') break;
	if(fname[i]=='.') {
	    append=false;
	    break;
	}
    }
    if(append) fname+=".la";

    ifstream* is = new ifstream(fname.c_str());
    if(is && (*is)) {
	openstream(is);
	return true;
    } else {
	delete is;
	return false;
    }
}

//
//  Kommando "read file" oder "read 'file.bla'" liest eine Datei ein
//  (a la #include)
//
bool read_func()
{
    string fname;

    while(1) {
        switch(ttype) {
         case TSYMBOL:
            fname += tstr;
            break;
         case TDOT:
            fname += ".";
            break;
         case TSLASH:
            fname += "/";
            break;
         case TSEMI:
            goto end_of_filename;
         default:
            error_message() << "Ungltiger Dateiname bei `read'." << endl;
            readeol();
            return false;
        }
        noalias_flag = true;
        getsymbol();
        noalias_flag = false;
    }

 end_of_filename:

    if(!do_read(fname)) {
	error_message() << "Datei kann nicht geffnet werden: " << fname << "." << endl;
    }
    return false;
}

//
//  Kommando "undef <macro>" lscht ein Makro
//
bool undef_func()
{
    if(ttype==TSYMBOL && tstr=="all") {
	macros_t::iterator i;
	for(i=macros.begin(); i!=macros.end();) {
	    delete i->second.contents;
	    macros.erase(i++);
	}
	true_ok = false;
	getsymbol();
	return true;
    }
    if(ttype!=TMACRO) {
	error_message() << "Nach `undef' mu ein Makroname oder `all' stehen." << endl;
	readeol();
	return false;
    }
    macros_t::iterator i = macros.find(tstr);
    if(i!=macros.end()) {
	if(tstr=="true") true_ok=false;
	delete i->second.contents;
	macros.erase(i);
    }
    getsymbol();
    return true;
}

//
//  I/O fr Reduktionsreihenfolge
//
struct TRedOrder {
    TReductionOrder o;
    const char* what;
};

TRedOrder red_orders[] = {
    { LAZY_REDUCTION,  "normal" },
    { BROAD_REDUCTION, "broad" },
    { EAGER_REDUCTION, "by-value" },
    { LAZY_REDUCTION,  "leftmost-outermost" },
    { LAZY_REDUCTION,  "by-name" },
    { LAZY_REDUCTION,  "lazy" },
    { LAZY_REDUCTION,  "normal-order" },
    { EAGER_REDUCTION, "eager" },
    { EAGER_REDUCTION, "leftmost-innermost" },
};

static bool order_name_in(void* ptr)
{
    TReductionOrder *optr = static_cast<TReductionOrder*>(ptr);
    if(ttype == TNUMBER) {
        if(tint > 0 && tint <= REDUCTIONORDER_MAX) {
            *optr = TReductionOrder(tint - 1);
            getsymbol();
            return true;
        }
    } else if(ttype == TSYMBOL) {
        for(size_t i = 0; i < countof(red_orders); i++)
            if(looking_at(red_orders[i].what)) {
                *optr = red_orders[i].o;
                getsymbol();
                return true;
            }
    }
    
    error_message() << "Ungltige Bezeichnung fr Auswerte-Strategie." << endl;
    if(readtty()) {
        cerr << "Gltige Werte sind:" << endl;
        for(size_t i = 0; i < countof(red_orders); i++)
            cerr << "\t" << red_orders[i].what << " (" << int(red_orders[i].o)+1 << ")" << endl;
    }
    return false;
}

static string order_name_out(void* ptr)
{
    TReductionOrder o = *static_cast<TReductionOrder*>(ptr);
    for(size_t i = 0; i < countof(red_orders); i++)
        if(red_orders[i].o == o) {
            return red_orders[i].what;
        }
    return "<error>";
}

//
//  bool-Flags I/O
//
static bool bool_flag_in(void* ptr)
{
    bool value;
    if((ttype==TNUMBER && tint==0) || (ttype==TSYMBOL && tstr=="off"))
 	value = false;
    else if((ttype==TNUMBER && tint==1) || (ttype==TSYMBOL && tstr=="on"))
 	value = true;
    else {
        error_message() << "Wert mu 0 oder 1 bzw. `on' oder `off' sein" << endl;
        return false;
    }
    *static_cast<bool*>(ptr) = value;
    getsymbol();
    return true;
}

static string bool_flag_out(void* ptr)
{
    return *static_cast<bool*>(ptr) ? "EIN" : "AUS";
}

static bool command_in(void* ptr);
static string command_out(void* ptr);

// Namen fr set
TSetName set_names[] = {
    { "alarm",        bool_flag_in, bool_flag_out, &alarm_flag,        "Signalton ausgeben, wenn die Auswertung beendet ist" },
    { "break-marker", bool_flag_in, bool_flag_out, &ctrlc_mark_flag,   "Abbruchpunkt bei `peval' markieren?" },
    { "default",      command_in,   command_out,   &default_command,   "Standard-Kommando (siehe Doku)" },
    { "eta",          bool_flag_in, bool_flag_out, &eta_flag,          "Eta-Reduktion durchfhren?" },
    { "freeform",     bool_flag_in, bool_flag_out, &freeform_flag,     "Freeform-Eingabe zulassen (Eingabe mit `;' beenden)" },
    { "full",         bool_flag_in, bool_flag_out, &full_paren_flag,   "Ausdrcke voll geklammert ausgeben" },
    { "inverse",      bool_flag_in, bool_flag_out, &inverse_flag,      "Prompt in `reverse video' ausgeben (ANSI/vt100)" },
    { "newlist",      bool_flag_in, bool_flag_out, &newlist_flag,      "Neues Listenformat (?nil.XX)" },
    { "number",       bool_flag_in, bool_flag_out, &number_flag,       "Ergebnisse als Zahlen auswerten" },
    { "post-alpha",   bool_flag_in, bool_flag_out, &post_alpha_flag,   "Ergebnis Alpha-Konvertieren?" },
    { "pre-alpha",    bool_flag_in, bool_flag_out, &pre_alpha_flag,    "Vorgezogene Alpha-Konversion?" },
    { "short-comb",   bool_flag_in, bool_flag_out, &magic_combinators, "Kurzschreibweise fr Kombinatoren (`Y' statt `<Y>')?" },
    { "ski-shortcut", bool_flag_in, bool_flag_out, &ski_shortcut_flag, "Verkrztes Verfahren fr SKI" },
    { "strategy",     order_name_in, order_name_out, &reduction_order, "Reduktions-Strategie" },
    { "timer",        bool_flag_in, bool_flag_out, &timer_flag,        "Zeit messen, die zur Reduktion bentigt wird" },
    { "verbose",      bool_flag_in, bool_flag_out, &verbose_flag,      "Statusmeldungen und Zwischenergebnisse anzeigen" },
    { "wait",         bool_flag_in, bool_flag_out, &wait_flag,         "Nach jedem Reduktionsschritt warten" },
    { 0, 0, 0, 0, 0 }
};

static void show_setting(TSetName* s)
{
    string val = s->outp(s->value);
    int indent = 18 - strlen(s->name) - val.length();
    if(indent <= 0)
        indent = 1;
    cout << s->name << " = " << val
         << string(indent, ' ') << s->help << endl;
}

//
//  Kommando "set" (Umgebung anzeigen)
//  Kommando "set name 0|1" (Umgebung ndern)
//
bool set_func()
{
    if(ttype == TSEMI) {
	for(int i = 0; set_names[i].name != 0; i++)
            show_setting(&set_names[i]);
	return true;
    }
    
    if(ttype != TSYMBOL) {
	error_message() << "`set' braucht 2 Argumente." << endl;
	readeol();
	return false;
    }
    
    string p = tstr;
    getsymbol();
    for(int i = 0; set_names[i].name; i++)
        if(p == set_names[i].name) {
            if(ttype == TSEMI) {
                show_setting(&set_names[i]);
                return true;
            } else {
                return set_names[i].func(set_names[i].value);
            }
        }
    
    error_message() << "unbekannte `set'-Einstellung." << endl;
    readeol();
    return false;
}

//
//  Kommando "echo term": Term voll geklammert ausgeben
//
bool echo_func()
{
    CNode* p = doreadlambda();
    if(!p) return false;
    
    bool fpf_save = full_paren_flag;
    full_paren_flag = true;
    
    p->print(cout);
    cout << endl;
    delete p;

    full_paren_flag = fpf_save;
    
    return true;
}

//
//  Kommando "fv term": Freie Variable bestimmen
//
bool fv_func()
{
    CNode* p = doreadlambda();
    if(!p) return false;
    TVariable* fv = p->freevars(0);
    if(!fv) {
	cout << "Keine freien Variablen." << endl;
    } else {
	for(TVariable* pv=fv; pv; pv=pv->next) {
	    cout << pv->name << endl;
	}
    }
    deletelist(fv);
    delete p;
    return true;
}

/* type oder annot-Kommando auf Ausdruck p */
void process_type(CNode* p, bool annotate)
{
    ctrlcpressed = false;

    int orig_type_nodes = CType::n_types;
    int o_unify = unify_count;
    bool show_term = false;
    time_t start = time(0);
    time_t end;
    int type_nodes;

    {
        /* Klammerblock, damit der CTypePtr bei der schlieenden Klammer
           zerstrt wird (fr Erkennung von Speicherlecks bei CTypes) */
        TTypeInfo ti;
        ti.annotate = annotate;
        
        CTypePtr ptr = p->get_type(ti, 0);
        cout.clear();

        end = time(0);
        
        if(ctrlcpressed) {
            cout << "Auswertung abgebrochen." << endl;
        } else if(ptr) {
            int rename_counter = 0;
            ptr->rename_unknown(rename_counter);
            cout << "Typ des Terms: " << ptr->to_string() << endl;
            
            bool first = true;
            for(map<string, CTypePtr>::iterator i = ti.free_types.begin();
                i != ti.free_types.end(); i++) {
                i->second->rename_unknown(rename_counter);
                cout << (first ? "mit " : "    ")
                     << i->first << " : " << i->second->to_string()
                     << endl;
                first = false;
            }
            show_term = annotate;
        } else {
            if(ti.error_msg.length())
                cout << "Term hat keinen Typ: " << ti.error_msg << endl;
        }
        type_nodes = CType::n_types - orig_type_nodes;
    }
    if(orig_type_nodes != CType::n_types && !annotate && !ctrlcpressed) {
        cerr << "Warnung: " << CType::n_types - orig_type_nodes
             << " Typ-Knoten briggeblieben (Programmfehler)" << endl;
    }
    if(show_term) {
	p->print(cout);
	cout << endl;
        pretty_print_result(p);
    }
    if(timer_flag) {
	cout << "Gesamtzeit: " << difftime(end, start)
             << " sec, " << unify_count - o_unify << " Unifikationen, "
             << type_nodes << " Typ-Knoten"
             << endl;
    }
}

//
//  Kommando "type|annot term": Term typisieren
//
bool type_helper(bool annot)
{
    if(ttype == TSYMBOL && tstr == "result") {
        getsymbol();
        if(!result)
            error_message() << "kein Ergebnis gespeichert." << endl;
        else
            process_type(result, annot);
    } else {
        CNode* p = doreadlambda();
        if(!p)
            return false;

        process_type(p, annot);
        delete p;
    }
    return true;
}

bool type_func()
{
    return type_helper(false);
}

bool annot_func()
{
    return type_helper(true);
}

//
//  Kommando "peval term": Term auswerten und gleichzeitig ausgeben
//
bool peval_func()
{
    CNode* p = doreadlambda();
    if(!p)
        return false;

    ctrlcpressed = abort_reduction = false;

    /*
     *  ist ntig, sonst `peval (?y x.x y) x (?x.x a)  =>*  (a a)'
     *  richtig wre     `eval (?y x.x y) x (?x.x a)   =>*  (x a)'
     */
    if(pre_alpha_flag)
        p->global_alpha(0, -1);
    
    TIoContext ioc;
    ioc.pc = PRINT_UNDEF;
    ioc.parens_needed = 0;
    int n = 0;
    while(!ctrlcpressed && !abort_reduction && p) {
	changed = false;
	CNode* q;
	q = p->ioreduce(cout, ioc);
        cout.clear();
        delete p;
	p = q;
	if(!changed)
            break;
        n++;
    }

    if(p) {
        if(ctrlc_mark_flag && (ctrlcpressed || abort_reduction)) {
            cout << "[Abbruch]";
        }
        p->print(cout, ioc.pc);
        while(ioc.parens_needed--)
            cout << ")";
    }
    cout << endl;
    if(timer_flag)
        cout << ((ctrlcpressed || abort_reduction) ? "Abbruch" : "Terminiert")
             << " nach " << n << " Schritten." << endl;
    return true;       
}

//
//  Kommando "show <macro>": Definition des Makros zeigen
//
bool show_func()
{
    bool fpf_save = full_paren_flag;
    if(ttype==TSYMBOL && tstr=="full") {
	getsymbol();
	full_paren_flag = true;
    } else if(ttype==TSYMBOL && tstr=="smart") {
	getsymbol();
	full_paren_flag = false;
    }
    
    if(ttype==TSYMBOL && (tstr=="all" || tstr=="ALL")) {
	bool really_all = (tstr=="ALL");
	macros_t::const_iterator i;
	for(i=macros.begin(); i!=macros.end(); i++)
	    if(really_all || i->second.doc.length()) {
		cout << "<" << i->first << "> = ";
		i->second.contents->print(cout);
		cout << endl << i->second.doc;
	    }
    } else {
	if(ttype!=TMACRO) {
	    error_message() << "`show' bentigt einen Makronamen oder `all' als Argument." << endl;
	    readeol();
	    full_paren_flag = fpf_save;
	    return false;
	}
	
	macros_t::iterator t = macros.find(tstr);
	if(t==macros.end())
	    cout << "Makro <" << tstr << "> ist nicht definiert." << endl;
	else {
	    t->second.contents->print(cout);
	    cout << endl << t->second.doc;
	}
    }
    full_paren_flag = fpf_save;
    getsymbol();
    return true;
}

bool ski_helper(CNode*& q)
{
    ctrlcpressed = false;
    q = 0;
    if(looking_at("result")) {
        getsymbol();
        if(!result) {
            error_message() << "Kein Ergebnis gespeichert." << endl;
            return true;
        }
        q = result->ski();
    } else {
        CNode* p = doreadlambda();
        if(!p)
            return false;
        q = p->ski();
        delete p;
    }
    cout.clear();
    return true;
}

//
//  Term in SKI-Form bringen
//
bool ski_func()
{
    CNode* q;
    bool b = ski_helper(q);
    if(q) {
        delete result;
        result = q;
        q->print(cout);
        cout << endl;
    }
    
    return b;
}

bool skieval_func()
{
    time_t starttime = time(0);
    time_t endtime = starttime;

    CNode* q;
    bool b = ski_helper(q);
    endtime = time(0);
    double convtime = difftime(endtime, starttime);
    
    delete result;
    starttime = time(0);
    result = 0;
    if(q) {
        ctrlcpressed = false;
        int n = 0;
        int old_i = i_reduce, old_k = k_reduce, old_s = s_reduce;
        do {
            changed = false;
            
            int kold_i = i_reduce, kold_k = k_reduce, kold_s = s_reduce;
            CNode* qq = q->ski_reduce();
            cout.clear();

            ++n;
            if(verbose_flag || wait_flag || !changed) {
                if(changed) {
                    cout << n;
                    char prefix = '[';
                    if(kold_s != s_reduce)
                        cout << prefix << s_reduce - kold_s << "S",
                            prefix = ',';
                    if(kold_k != k_reduce)
                        cout << prefix << k_reduce - kold_k << "K",
                            prefix = ',';
                    if(kold_i != i_reduce)
                        cout << prefix << i_reduce - kold_i << "I",
                            prefix = ',';
                    if(prefix != '[')
                        cout << "]: ";
                    else
                        cout << ": ";
                }
                qq->print(cout);
                after_eval();
            }
            delete q;
            q = qq;
        } while(changed && !ctrlcpressed);
        if(timer_flag) {
            endtime = time(0);
            cout << "Gesamtzeit: " << difftime(endtime, starttime) << " sec. (+ "
                 << convtime << " Konv.), "
                 << n << " Schritte ("
                 << s_reduce - old_s << "S + "
                 << k_reduce - old_k << "K + "
                 << i_reduce - old_i << "I = "
                 << s_reduce - old_s + k_reduce - old_k + i_reduce - old_i
                 << ")" << endl;                
        }
        result = q;
    }
    return b;
}

//
//  Manual anzeigen
//
bool manual_func()
{
    const char* p = getenv("PAGER");
    if(!p)
        p = "less";
    string command = string(p) + " " DIRECTORY "/lambda.txt";
    int i = system(command.c_str());
    if(i != 0) {
	cerr << "Fehler bei der Ausfhrung des Kommandos `" << command << "'" << endl;
    }
    return false;
}

//
//  Systemstatus
//
bool info_func()
{
    int mcount=0, dcount=0, acount=0;
    for(macros_t::const_iterator i=macros.begin(); i!=macros.end(); i++) {
	mcount++;
	if(i->second.doc.length()) dcount++;
	acount += i->second.contents->count_atoms();
    }
    
    cout << "System-Information - Lambda Version " VERSION << endl
	 << endl
	 << "main.cc compiliert am        : " __DATE__ " um " __TIME__ << endl
#ifdef WITH_READLINE
	 << "Readline Version             : " << rl_library_version << endl
	 << "Readline Application Name    : " << rl_readline_name << endl
#endif
	 << "Anzahl bekannte Makros       : " << mcount << endl
	 << "- davon dokumentiert         : " << dcount << endl
	 << "- enthaltene Knoten          : " << acount << endl
         << "Typknoten im System          : " << CType::n_types << endl;

    if(result)
	cout << "Knoten im akt. Ergebnis      : " << result->count_atoms() << endl;
    cout << "Anonyme Variablen (a-Konv.)  : " << tempnum << endl
         << "Anzahl Beta-Reduktionen ()  : " << n_beta << endl
         << "Anzahl Eta-Reduktionen (n)   : " << n_eta << endl
         << "Anzahl Makro-Expansionen (e) : " << n_expand << endl;
    if(s_reduce || i_reduce || k_reduce)
        cout << "Anzahl S-Reduktionen (S)     : " << s_reduce << endl
             << "Anzahl K-Reduktionen (K)     : " << k_reduce << endl
             << "Anzahl I-Reduktionen (I)     : " << i_reduce << endl;

    return true;
}

// Sprungtabelle fr Kommandos
bool help_func();
  
TCommand commands[] = {
    { "eval", eval_func, true,	    "eval term           - lambda-Term reduzieren" RL_COMPLETE(0) },
    { "peval", peval_func, true,    "peval term          - lambda-Term reduzieren und ausgeben gleichzeitig" RL_COMPLETE(0) },
    { "def", define_func, true,	    "def <name> term     - Makro definieren" RL_COMPLETE(0) },
    { "edef", edef_func, true,	    "edef <name> term    - Makro definieren (Term expandieren)" RL_COMPLETE(0) },
    { "rdef", rdef_func, true,	    "rdef <name> term    - Makro definieren (Makro-Rekursion entfernen)" RL_COMPLETE(0) },
    { "undef", undef_func, true,    "undef <name>|all    - Makro lschen" RL_COMPLETE(0) },
    { "read", read_func, true,	    "read datei          - Kommandos aus Datei lesen" RL_COMPLETE(la_file_completion_function) },
    { "set", set_func, true,	    "set [option wert]   - Einstellung ndern / anzeigen" RL_COMPLETE(complete_options) },
    { "help", help_func, false,	    "help                - diesen Text anzeigen" RL_COMPLETE(0) },
    { "manual", manual_func, false, "manual              - Anleitung (lambda.txt) anzeigen" RL_COMPLETE(0) },
    { "echo", echo_func, true,	    "echo term           - Term vollstndig geklammert ausgeben" RL_COMPLETE(0) },
    { "fv", fv_func, true,	    "fv term             - freie Variablen ermitteln" RL_COMPLETE(0) },
    { "show", show_func, true,	    "show <name>|all     - Makrodefinition anzeigen" RL_COMPLETE(0) },
    { "result", result_func, true,  "result              - letztes (Zwischen-)Ergebnis anzeigen" RL_COMPLETE(0) },
    { "cont", cont_func, true,      "cont                - Zwischenergebnis weiter auswerten" RL_COMPLETE(0) },
    { "type", type_func, true,      "type term|result    - Typ von Term ermitteln" RL_COMPLETE(0) },
    { "annot", annot_func, true,    "annot term|result   - Typ von Term ermitteln und Term kommentieren" RL_COMPLETE(0) },
    { "ski", ski_func, true,        "ski term|result     - Term in SKI-Form bringen (Abstraktions-Elimination)" RL_COMPLETE(0) },
    { "skieval", skieval_func, true, "skieval term|result - Term im SKI-System auswerten" RL_COMPLETE(0) },
    { "info", info_func, true,      "info                - Information zum Systemstatus" RL_COMPLETE(0) },
    { 0, 0, false, 0 RL_COMPLETE(0) }
};

bool command_in(void* ptr)
{
    int* cmd = static_cast<int*>(ptr);
    if(ttype == TSYMBOL) {
        int i = 0;
        while(commands[i].name) {
            if(commands[i].name == tstr) {
                if(!commands[i].doget) {
                    error_message() << "`" << tstr << "' kann nicht mit `set default' verwendet werden" << endl;
                    return false;
                } else {
                    getsymbol();
                    *cmd = i;
                    return true;
                }
            }
            ++i;
        }
    }
    error_message() << "Kommandowort erwartet nach `set default'" << endl;
    return false;
}

string command_out(void* ptr)
{
    return commands[*static_cast<int*>(ptr)].name;
}
    
//
//  Kommando "help"
//
bool help_func()
{
    readeol();
    cout << "Stefan's 100% Pure Lambda Program - Version " VERSION << endl
	 << endl
	 << "Abstraktion: `?x.y,' `\\x.y' oder `lambda x.y', Abkrzung `?x y.X = ?x.?y.X" << endl
	 << "Applikation: linksassoziative Klammerung `x y z', rechtsassoziativ `x/y/z'" << endl
	 << "Zahlen- (`12') und Listenreprsentation (`[a b c]'). Case-sensitiv. Makronamen" << endl
	 << "in spitzen Klammern: `<succ>'. Typen: `?x:typ.y', typ=Symbol oder `typ->typ'." << endl
	 << endl;

    int i = 0;
    for(TCommand* p=commands; p->name!=0; p++, i++)
	cout << p->help << (i == default_command ? " [*]" : "") << endl;

    cout << "exit                - Programmende" << endl
	 << endl;
    if(freeform_flag)
	cout << "Anweisungen mssen mit Semikolon enden!" << endl;
    cout << "Wichtig: RTFM. Das FM heit `lambda.txt' (Kommando `manual')." << endl;
    return false;
}

///////////////////////////////////////////////////////////////
//
//   Hauptprogramm
//
///////////////////////////////////////////////////////////////
int main()
{
#ifdef WITH_SIGACTION
    /* wir wollen: bei C-c den ctrlchandler aufrufen, Systemaufrufe
     * dadurch nicht unschdlich machen und keine anderen Signale
     * blockieren */
    set.sa_handler = ctrlchandler;
    sigemptyset(&set.sa_mask);
    set.sa_flags = SA_RESTART;

    if(signal(SIGINT,SIG_IGN)!=SIG_IGN) {
	sigaction(SIGINT,&set,&get);
    }
#else
    /* allerdings ist ANSI signal auch nicht schlecht... */
    if(signal(SIGINT, SIG_IGN) != SIG_IGN)
        signal(SIGINT, ctrlchandler);
#endif

    const char* emacs = getenv("EMACS");
    if(emacs && strcmp(emacs, "t") == 0)
        inverse_flag = 0;

#ifdef WITH_READLINE
    init_readline();
#endif

    cout << "Stefan's 100% Pure Lambda - Version " VERSION << endl
	 << "(c) copyright 1998,1999,2000 by Stefan Reuther <Streu@gmx.de>" << endl
	 << endl
	 << "? x.y   = lambda x.y" << endl
	 << "help    = Hilfe" << endl
	 << "exit    = Ende" << endl
	 << endl;

    startscan();

    /*
     *  AAAAAAAARRRRRRRRRGGGGGGHHHHH!
     *
     *  do_read ist "non-blocking". Dieser Code fhrt also dazu, da
     *  _erst_ die Standardbibliothek, und dann die autoexec geladen
     *  werden (quasi wie ein `read stdlib'-Kommando in `autoexec.la').
     */
    if(!do_read("autoexec.la")) {
        const char* c = getenv("HOME");
        if(c) {
            std::string s = c;
            s += "/autoexec.la";
            do_read(s);
        }
    }
    
    if(!do_read(DIRECTORY "/stdlib")) {
	cout << "Warnung: Standardbibliothek kann nicht geladen werden." << endl;
    }

    while(1) {
	resetscan();
	if(inverse_flag) prompt = INV_ON "lambda>" INV_OFF " ";
	else prompt = "lambda> ";
		
	getsymbol();

	if(inverse_flag) prompt = INV_ON "      >" INV_OFF " ";
	else prompt = "      > ";
	int cmd = default_command;
	if(ttype==TSEMI) {
	    continue;
	} else
	    if(ttype==TSYMBOL) {
		if(tstr=="exit")
                    break;
		for(int i=0; commands[i].name!=0; i++) {
		    if(tstr==commands[i].name) {
			cmd=i;
			if(commands[i].doget)
                            getsymbol();
			break;
		    }
		}
	    } else if(ttype==TEOF)
                break;
	if((*commands[cmd].func)()) {
	    if(ttype!=TSEMI) {
		error_message() << "Semikolon fehlt." << endl;
		readeol();
	    }
	}
    }
    cout << endl;
}
