//
//  Lambda
//  . Global benutzte Dinge
//  . CNode (Basisklasse)
//  . CLambda (Reprsentation von lambda x.term)
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//

#include <string>
#include <iostream>
#include <cstdio>
#include "node.h"

int tempnum = 0;                // Nummer fr temp-Variablen
bool changed;                   // letztes reduce hat was gebracht?
bool full_paren_flag = false;   // voll geklammerte Ausgabe
bool newlist_flag = true;       // neues Listenformat (?nil. ...)
bool true_ok = false;           // <true> definiert und ok?
bool pre_alpha_flag  = false;   // Alphakonversion vorher durchfhren?
bool post_alpha_flag = false;   // Ausgabe alpha-konvertieren?
bool ski_shortcut_flag = true;

TReductionOrder reduction_order;// Auswertereihenfolge

int n_eta = 0, n_beta = 0, n_expand = 0;

extern bool eta_flag;          // main.cc

//
//  Liefert Namen fr neue Temporrvariable fr Alpha-Konversion
//
string newtemp()
{
    tempnum++;
    std::string s = "_";
    int i=tempnum;
    while(i>0) {
	s += (char)('0' + i%10);
	i /= 10;
    }
    return s;
}

/***** CNode *****/

CNode* CNode::ski_reduce()
{
    return copy();
}

CNode* CNode::ioreduce(std::ostream&, TIoContext& ioc)
{
    return reduce();
}

bool CNode::calls_macro(std::string)
{
    return false;
}

CNode* CNode::replace_macro(std::string m, std::string s)
{
    return this;
}

/***** CLambda *****/

//
//  Liefert neuen Baum mit reduziertem Ausdruck, der Original-Ausdruck
//  wird invalidiert.
//
//  lambda x.y wird nicht beta-reduziert(!). Das geschieht
//  in CApplication ber transform()
//
CNode* CLambda::reduce()
{
    register CNode* eta = 0;

    // Eta-Reduktion, falls mglich
    if(eta_flag)
        eta = sub->eta(0, symbol);

    if(eta) {
        changed = true;
        n_eta++;
        return eta;
    } else
        return new CLambda(symbol, do_reduce(sub), type);
}

CNode* CLambda::copy()
{
    return new CLambda(symbol, sub->copy(), type);
}

void CLambda::print(std::ostream& out, TPrintContext pc)
{
    if(full_paren_flag) {
	out << "(?" << symbol;
        if(type)
            out << ":" << type->to_string();
        out << ".";
	sub->print(out);
	out << ")";
    } else {
	if(pc!=PRINT_LAMBDA) out << "(?";
	out << symbol;
        if(type)
            out << ":" << type->to_string();
	if(sub->lambda()) {
	    out << " ";
	    sub->print(out,PRINT_LAMBDA);
	} else {
	    out << ".";
	    sub->print(out,PRINT_ABSTRACT);
	}
	if(pc!=PRINT_LAMBDA) out << ")";
    }
}

//
//  Die Variablen in /p/ sind frei. Packe die freien Variablen aus
//  diesem Zweig *vor* /p/ und liefere neuen Listen-Anfangspunkt.
//
TVariable* CLambda::freevars(TVariable *p)
{
    WHERE;
    TVariable *q, *ql;

    /* suche die von sub gebundenen Variablen */
    if(p)
        p->marker=true;  // Die Markierung gibt an, welche
                         // Variablen "vorher" schon frei waren
                         // und von hier nicht gebunden werden knnen
    if(sub)
        p = sub->freevars(p);

    /* lsche "diese" Variable, wenn sie nicht schon vorher frei war */
    q=p;
    ql=0;
    while(q && !q->marker) {
	if(q->name == symbol) {
	    if(!ql) p=q->next;
	    else ql->next=q->next;
	    delete q;
	    break;
	}
	ql=q;
	q=q->next;
    }
    return p;
}

//
//  Alphakonversion: sorge dafr, da alle Variablen in /fv/ nicht
//  mehr gebunden sind. CLambda weist die Ersetzvariablen (replaceby)
//  zu.
//
bool CLambda::alpha(TVariable *fv)
{
    if(!sub) return false;
	
    TVariable* q = fv;
    while(q && (q->name!=symbol))
        q=q->next;
    
    if(q) {
	std::string old = q->replaceby;
	symbol = q->replaceby = newtemp();
	bool res=sub->alpha(fv);
	q->replaceby = old;
	return res;
    }
    return sub->alpha(fv);
}

//
//  liefert einen lesbaren Namen korrespondierend zu slot n
//
static std::string get_user_name(int n)
{
    char tmp[20];
    char c = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"[n % 52];
    int x = n / 52;
    if(x) {
        std::sprintf(tmp, "%c%d", c, x);
        return tmp;
    } else
        return string(1, c);
}

//
//  true iff die Variable /s/ wird bereits als Ersatztext fr eine
//  Variable in /fv/ verwendet
//
static bool is_used(TVariable* p, std::string s)
{
    while(p) {
        if(p->replaceby == s)
            return true;
        p = p->next;
    }
    return false;
}

//
//  Liefert einen Variablennamen, der nicht mit einer der
//  freien Variablen kollidiert. Setzt /id/ entsprechend.
//
static std::string get_free_user_name(int& id, TVariable* free)
{
    std::string s;
 try_again:
    s = get_user_name(id++);
    for(TVariable* p = free; p != 0; p = p->next)
        if(p->name == s)
            goto try_again;
    return s;
}

//
//  "Great Renaming"
//
//  Variablen umbenennen entsprechend der Liste /fv/ und der
//  angegebenen /action/:
//  action >= 0: anonyme Variablen benamsen
//  action < 0:  alles anonymisieren
//
void CLambda::global_alpha(TVariable* fv, int action)
{
    if(!sub)
        return;

    if(action < 0) {
        /* action < 0: anonymisieren */
        TVariable q(fv, symbol);
        q.replaceby = symbol = newtemp();
        sub->global_alpha(&q, action);
    } else {
        /* action > 0: personalisieren */
        if(symbol[0] == '_' || is_used(fv, symbol)) {
            /* anonyme bzw. bereits vergebene Variablen umbenennen:
               ?_1 a b.foo => ?a b c.foo */
            TVariable q(fv, symbol);

            TVariable* free = freevars(0);
            q.replaceby = symbol = get_free_user_name(action, free);
            deletelist(free);
            
            sub->global_alpha(&q, action);
        } else
            sub->global_alpha(fv, action);
    }
}

//
//  Ersetze Variable /sym/ durch /rep/. Liefert neuen Baum.
//
CNode* CLambda::replace(std::string& sym, CNode* rep)
{
    /* /sym/ wird neu gebunden -- ignorieren */
    if(sym==symbol)
//         return copy();
        return new CLambda(symbol, move(sub), type);
    return new CLambda(symbol, do_replace(sub,sym,rep), type);
}

//
//  Wird von CApplication::reduce aufgerufen:
//    (?x.t) a  =>  (?x.t)->transform(a)
//
CNode* CLambda::transform(CNode* rep)
{
    return do_replace(sub, symbol, rep);
}

bool CLambda::numcheck(TNumCheck& nc)
{
    if(nc.status==TNumCheck::CHECK_X) {	// ?x.?y.(bla)
	nc.x = symbol;			// ^^^
	nc.status = TNumCheck::CHECK_Y;
	return sub->numcheck(nc);
    } else if(nc.status==TNumCheck::CHECK_Y) {	// ?x.?y.(bla)
	nc.y = symbol;                          //    ^^^
	nc.status = TNumCheck::CHECK_RIGHT;
	return sub->numcheck(nc);
    } else
	return false;
}

bool CLambda::listcheck(TListCheck& lc)
{
    if(lc.status==TListCheck::CHECK_TAIL) {       // ?f.f a (?f.f b nil)
	lc.f = symbol;                            // ^^^^
	lc.status = TListCheck::CHECK_APP;
	lc.length++;
	return sub->listcheck(lc);
    } else if(lc.status==TListCheck::CHECK_NIL) { // ?nil.?f.f a (..nil)))
	lc.nil = symbol;                          // ^^^^^
	lc.status = TListCheck::CHECK_TAIL;
	return sub->listcheck(lc);
    } else
	return lc.status==TListCheck::CHECK_2ND;  // ?f.f a (?f.f b nil)
}                                                 //      ^

void CLambda::listout(std::ostream& o, int n, char delim)
{
    if(n==TListCheck::CHECK_TAIL) {
	sub->listout(o,TListCheck::CHECK_APP, delim);
    } else if(n==TListCheck::CHECK_NIL) {
	sub->listout(o,TListCheck::CHECK_TAIL, delim);
    }
}

CTypePtr CLambda::get_type(TTypeInfo& ti, TTypeItem* local_types)
{
    /*
     *   type(?x.M) = type(x) -> type(M)
     */
    TTypeItem my_type;
    my_type.next = local_types;
    my_type.name = symbol;
    if(type)
        my_type.type = type;
    else {
        my_type.type = new CType();
        if(ti.annotate)
            type = my_type.type;
    }

    CTypePtr m = sub->get_type(ti, &my_type);
    if(m)
        return new CType(my_type.type, m);
    else
        return 0;
}

CNode* CLambda::ski()
{
    if(eta_flag) {
        CNode* b = sub->eta(0, symbol);
        if(b) {
            CNode* a = b->ski();
            delete b;
            return a;
        }
    }    
    CNode* a = sub->ski();
    if(a) {
        CNode* r = a->ski_lambda(symbol);
        delete a;
        return r;
    } else
        return 0;
}

CNode* CLambda::ski_lambda(std::string s)
{
    std::cerr << "Fehler: CLambda::ski_lambda() aufgerufen\n";
    std::exit(1);
}

//
//  Lscht die Variablenliste pv
//
void deletelist(TVariable* pv)
{
    while(pv) {
	TVariable* p = pv;
	pv = pv->next;
	delete p;
    }
}

//
//  Liste ausgeben
//
bool try_print_list(std::ostream& o, CNode* p, bool report)
{
    TListCheck lc;
    TListCheck::TStatus starter;
    if(newlist_flag)
	lc.status = starter = TListCheck::CHECK_NIL;
    else {
	lc.status = starter = TListCheck::CHECK_TAIL;
	lc.nil = "nil";
    }
    lc.length = 0;
    if(p->listcheck(lc)) {
	if(report)
            o << "Liste der Lnge " << lc.length << endl;
	if(lc.length)
            p->listout(o,starter,'[');
	else o << "[";
	o << "]";
	return true;
    }
    return false;
}

CNode* CLambda::ioreduce(std::ostream& os, TIoContext& ioc)
{
    register CNode* eta = 0;

    // Eta-Reduktion, falls mglich
    if(eta_flag)
        eta = sub->eta(0, symbol);

    if(eta) {
        changed = true;
        n_eta++;
        return eta;
    } else {
        // keine Eta-Reduktion
        os << "(?" << symbol;
        if(type)
            os << ":" << type->to_string();
        os << "." << flush;

        ioc.pc = PRINT_ABSTRACT;
        ioc.parens_needed++;
        CNode* q = sub->ioreduce(os, ioc);
        delete sub;
        sub = 0;
        return q;
    }
}

bool CLambda::calls_macro(std::string s)
{
    return sub->calls_macro(s);
}

CNode* CLambda::replace_macro(std::string mac, std::string id)
{
    sub = sub->replace_macro(mac, id);
    return this;
}
