//
//  Lambda: Kurze Zahlen-Schreibweise
//  . CNumber
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//
#include <iostream>
#include <string>
#include "number.h"
#include "node.h"
#include "symbol.h"
#include "application.h"

CNode* CNumber::reduce()
{
    if(!expanded) {
        makevalue();
        if(expanded)
            return move(expanded);
    }
    if(expanded)
        return do_reduce(expanded);
    else
        return copy();
}

CNode* CNumber::eta(int stage, std::string s)
{
    if(!expanded)
        return 0;
    return expanded->eta(stage, s);
}

CNode* CNumber::copy()
{
    if(expanded)
        return expanded->copy();
    else
        return new CNumber(val, 0);
}

void CNumber::print(std::ostream& out, TPrintContext pc)
{
    if(expanded) expanded->print(out, pc);
    else out << val;
}

CLambda* CNumber::lambda()
{
    if(!expanded) return 0;
    return expanded->lambda();
}

TVariable* CNumber::freevars(TVariable* p)
{
    if(!expanded) return p;
    return expanded->freevars(p);
}

bool CNumber::alpha(TVariable* fv)
{
    if(expanded) return expanded->alpha(fv);
    return false;
}

void CNumber::global_alpha(TVariable* fv, int action)
{
    if(expanded)
        expanded->global_alpha(fv, action);
}

void CNumber::makevalue()
{
    if(expanded || val<0) return;

    std::string x = "x", y = "y";
    if(pre_alpha_flag)
        x = newtemp(),
        y = newtemp();
    
    CNode* p = new CSymbol(y);
    for(int i=0; i<val; i++) {
	p = new CApplication(new CSymbol(x), p);
    }
    expanded = new CLambda(x, new CLambda(y, p, 0), 0);
    n_expand++;
    changed = true;
}

CNode* CNumber::replace(std::string& sym, CNode* rep)
{
    if(expanded) return do_replace(expanded,sym,rep);
    else return copy();
}

// nur bereits expandierte Zahlen werden als solche erkannt
bool CNumber::numcheck(TNumCheck& nc)
{
    if(expanded) return expanded->numcheck(nc);
    return false;
}

CTypePtr CNumber::get_type(TTypeInfo& ti, TTypeItem* local_types)
{
    if(!expanded)
        makevalue();
    if(!expanded) {
        /* should not happen */
        ti.error_msg = "Fehler beim expandieren einer Zahl";
        return 0;       
    }
    return expanded->get_type(ti, local_types);
}

CNode* CNumber::ski()
{
    if(!expanded)
        makevalue();

    return expanded->ski();
}

CNode* CNumber::ski_lambda(std::string s)
{
    if(!expanded)
        makevalue();

    return expanded->ski_lambda(s);
}
