//
//  Lambda
//  . Parser
//
//  (c) copyright 2000 by Stefan Reuther <Streu@gmx.de>
//

#include <string>
#include <iostream>
#include <cctype>

#include "node.h"
#include "type.h"
#include "scanner.h"
#include "symbol.h"
#include "macro.h"
#include "application.h"
#include "number.h"

using namespace std;

bool magic_combinators = true;

CNode* readlambda();
CNode* readlist();
CTypePtr readtype();

//
//  true iff der angegebene Bezeichner ist "magisch"
//
bool magic_id(string s)
{
    return s.length() == 1 && isupper(s[0]);
}

//
//  Funktionstyp (mglicherweise zusammengesetzt):
//  `type' oder `type->functype'
//
CTypePtr readfunctype()
{
    CTypePtr ret = readtype();
    if(ttype == TARROW) {
        getsymbol();
        return new CType(ret, readfunctype());
    } else
        return ret;
}

//
//  Typ-"Atom": `symbol' oder `( functype )'
//
CTypePtr readtype()
{
    if(ttype == TSYMBOL) {
        CTypePtr ret = new CType(tstr);
        getsymbol();
        return ret;
    } else if(ttype == TLEFT) {
        getsymbol();
        CTypePtr ret = readfunctype();
        if(ttype != TRIGHT)
            throw string("`)' erwartet nach Typangabe.");
        getsymbol();
        return ret;
    } else
        throw string("Ungltige Typangabe.");
}

//
//  "bla.fasel" in irgendeinem "lambda fred bla.fasel" einlesen
//  (abgekrzte Schreibweise)
//
CNode* readlambdaseq()
{
    if(ttype != TSYMBOL)
        throw string("Erwarte ein Symbol nach lambda.");
    string s = tstr;
    CNode* p;
    CTypePtr type = 0;

    if(getsymbol() == TCOLON) {
        getsymbol();
        type = readfunctype();
    }
    
    if(ttype == TSYMBOL) {
	p = readlambdaseq();
    } else if(ttype != TDOT) {
	throw string("Der Punkt nach den lambda-Variablenbindungen fehlt.");
    } else {
	getsymbol();
	p = readlambda();
    }
    if(p)
        return new CLambda(s, p, type);
    else
        throw string("Nach `.' mu ein nicht-leerer Term folgen.");
}

//
//  Primitive == irgendetwas auer der einfachen ungeklammerten
//  Applikation (->Listenelement)
//
CNode* readprimitive()
{
    register CNode* p;
    if(ttype==TLEFT) {
	getsymbol();
	p = readlambda();
	if(ttype!=TRIGHT)
            throw string("`)' fehlt.");
	if(!p)
            throw string("Zwischen Klammern mu ein nicht-leerer Term stehen.");
	getsymbol();
    } else if(ttype==TSYMBOL) {
        if(magic_combinators && magic_id(tstr) && macros.find(tstr) != macros.end())
            p = new CMacro(tstr);
        else
            p = new CSymbol(tstr);
	getsymbol();
    } else if(ttype==TMACRO) {
	p=new CMacro(tstr);
	getsymbol();
    } else if(ttype==TNUMBER) {
	p=new CNumber(tint);
	getsymbol();
    } else if(ttype==TLEFTSQR) {
	getsymbol();
	p=readlist();
	if(newlist_flag) {
	    p=new CLambda("nil", p, 0);
	}
    } else {
	return 0;
    }
    return p;
}

//
//  Listenreprsentation
//
CNode* readlist()
{
    register CNode *p, *q;
    if(ttype==TRIGHTSQR) {
	p=new CSymbol("nil");
	getsymbol();
    } else {
	p = readprimitive();
	if(!p)
            throw string("Erwarte einfachen Ausdruck innerhalb einer Liste.");
	try {
	    q = readlist();
	}
	catch(...) {
	    delete p;
	    throw;
	}

	string argname = newtemp();  // vermeidet Kollisionen mit Element 'f'

	p = new CApplication(new CSymbol(argname), p);
	p = new CApplication(p, q);
	return new CLambda(argname, p, 0);
    }
    return p;
}

//
//  Lambda-Term einlesen
//  (Abstraktion, Applikation, Makro, Zahl, Symbol)
//
CNode* readlambda()
{
    CNode *p=0, *q=0;
	
    try {
	while(1) {
	    if((p=readprimitive())) {
	    } else if(ttype==TSLASH) {
		if(!q)
                    throw string("Erwarte nicht-leeren Term vor `/'.");
		getsymbol();
		p = readlambda();
		if(!p)
                    throw string("Erwarte nicht-leeren Term nach `/'.");
	    } else if(ttype==TLAMBDA) {
		getsymbol();
		p=readlambdaseq();
	    } else {
		return q;
	    }

	    if(q) q = new CApplication(q,p);
	    else q = p;
	}
    }
    catch(...) {
	delete q;
	throw;
    }
}

//
//  Wrapper fr readlambda(): fngt Parserfehler und gibt ggf. NULL
//  zurck.
//
CNode* doreadlambda()
{
    try {
	return readlambda();
    }
    catch(string& s) {
	readeol();
	error_message();
	cerr << s << endl;
	return 0;
    }
}

