//
//  Lambda - Readline Schnittstelle
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//

#ifdef WITH_READLINE
#define _POSIX_SOURCE

#include "rl-stuff.h"
#include "macro.h"
#include "main.h"
#include <cstdlib>
#ifdef WITH_DIRS
#include <sys/types.h>
#include <dirent.h>
#endif

//
//  Komplettiert Makronamen
//
char* macro_completion_function(char* text, int state)
{
    static macros_t::iterator i;

    if(*text != '<') return 0;
    text++;
	
    if(!state) {
	i = macros.begin();
    }

    int n = strlen(text);
    while(1) {
	if(i==macros.end()) return 0;
	if(n==0 || strncmp(i->first.c_str(),text,n)==0) break;
	i++;
    }

    char* p = (char*)malloc(i->first.length()+3);
    if(!p) return 0;

    strcpy(p,"<");
    strcat(p,i->first.c_str());
    strcat(p,">");
    i++;
    return p;
}

char* complete_options(char* text, int state)
{
    static TSetName* p;
    
    if(!state) p=set_names;

    while(p->name) {
	if(*text==0 || strncmp(p->name,text,strlen(text))==0) {
	    char* result = (char*)malloc(strlen(p->name)+1);
	    if(result) strcpy(result,p->name);
	    p++;
	    return result;
	}
	p++;
    }
    return 0;
}

//
//  Komplettiert Dateinamen
//
char* la_file_completion_function(char* text, int state)
{
#ifndef WITH_DIRS
    return 0;
#else
    static DIR* dir=0;

    if(!state) {
	if(dir) closedir(dir);
	dir = opendir(".");
	if(!dir) return 0;
    }

    while(1) {
	dirent* ent = readdir(dir);
	if(!ent) break;

	size_t i = strlen(ent->d_name);
	if(i>=3+strlen(text) && ent->d_name[i-1]=='a' && ent->d_name[i-2]=='l' && ent->d_name[i-3]=='.'
	   && (*text==0 || strncmp(text,ent->d_name,strlen(text))==0))
	{
	    char* c = (char*)malloc(i+1);
	    if(c) {
		strcpy(c,ent->d_name);
		c[i-3]=0;
	    }
	    return c;
	}
    }
    closedir(dir);
    dir=0;
    return 0;
#endif    
}

//
//  Komplettiert Kommandonamen
//
char* command_completion_function(char* text, int state)
{
    static TCommand* p;
    char* result;

    if(!state) {
	p=commands;
	result="exit";
    } else {
	result = (*p++).name;
    }

    while(1) {
	if(!result) return 0;
	if(*text==0 || strncmp(result,text,strlen(text))==0) break;
	result = (*p++).name;
    }

    char* q = (char*)malloc(strlen(result)+1);
    if(!q) return 0;
    strcpy(q,result);
    return q;
}

//
//  Allgemeine Komplettierungs-Funktion
//
char** lambda_completer(char* text, int start, int end)
{
    char** matches = 0;

    bool command_position = true;
    int i=start-1;
    if(i>=0) {
	while(i>=0 && (rl_line_buffer[i]==' ' || rl_line_buffer[i]=='\t'))
	    i--;
	if(i>=0 && rl_line_buffer[i]!=';') command_position = false;
    }

    // Makronamen
    if(*text == '<')
	matches = completion_matches(text, macro_completion_function);

    // Kommandos
    if(!matches && command_position)
	matches = completion_matches(text, command_completion_function);

    // `custom completers' fr einzelne Kommandos
    if(!matches && !command_position) {
	// suche Beginn des Kommandos
	while(i>=0 && rl_line_buffer[i]!=';') i--;
	do {
	    i++;
	} while(i<start && (rl_line_buffer[i]==' ' || rl_line_buffer[i]=='\t'));

	if(i<start) {
	    for(int cmd=0; commands[cmd].name!=0; cmd++)
                /* Cast beseitigt eine Warnung; Kommandonamen sind kurz genug */
		if(i + (int) strlen(commands[cmd].name) < start
		   && strncmp(rl_line_buffer+i, commands[cmd].name, strlen(commands[cmd].name))==0)
		{
		    if(commands[cmd].completer!=0)
			matches = completion_matches(text, commands[cmd].completer);
		    break;
		}
	}
    }

    if(matches && !*matches) {
	free(matches);
	matches=0;
    }

    return matches;
}

//
//  verhindert filename completion
//
char* just_say_no(char* text, int state)
{
    return 0;
}

//
//  Info ber Makro (Alt-I)
//
int describe_macro(int count, int key)
{
    int point = rl_point;
    char* p = rl_line_buffer;

    if(point>=rl_end) point--;
    while(point>=0 && p[point]==' ') point--; /* automagisch eingefgter Leerraum */
                                               /* nach Completion */
    int xpoint = point;
    while(point>=0 && p[point]!='<') {
	if(p[point]==' ' || p[point]=='\t') {
	    ding();
	    return 0;
	}
	point--;
    }
    if(point<0) {
	ding();
	return 0;
    }

    int endp = point;
    while(endp<=rl_end && p[endp]!='>') {
	if(p[endp]==' ' || p[endp]=='\t') {
	    ding();
	    return 0;
	}
	endp++;
    }
    if(endp>rl_end || endp<xpoint-1) {
	ding();
	return 0;
    }

    p += point+1;
    std::string s(p, endp-point-1);

    macros_t::iterator i = macros.find(s);
    if(i==macros.end()) {
	ding();
	return 0;
    }
    if(i->second.doc.length())
	std::cout << std::endl
                  << "Makro <" << s << ">:" << std::endl
                  << i->second.doc;
    else
	std::cout << std::endl
                  << "Makro <" << s << "> ist nicht dokumentiert." << std::endl;
    
    rl_reset_line_state();
    return 0;
}

//
//  Initialisiert Readline
//
void init_readline()
{
    using_history();
    rl_readline_name = "lambda";
    rl_completion_entry_function = just_say_no;
    rl_attempted_completion_function = lambda_completer;
    rl_basic_word_break_characters = " \t\n\"';[]()./";
    rl_add_defun("describe-macro",describe_macro,-1);
    rl_bind_key_in_map('i',describe_macro,emacs_meta_keymap);
}
#endif
