//
//  Lambda - Scanner
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//

#include <iostream>
#include <string>
#include <cctype>
#include "scanner.h"
#include "rl-stuff.h"

using namespace std;

// Stack aller offenen Dateien (vom read Kommando)
struct TIncludeStack {
    TIncludeStack *prev;
    istream* str;
    int prevc;
    int lineno;
	
    TIncludeStack(TIncludeStack* aprev, istream* astr, int aprevc)
	: prev(aprev), str(astr), prevc(aprevc), lineno(1) {}
};

TType ttype;		// aktueller Tokentyp
string tstr;		// aktueller Tokenname
int tint;		// aktueller Wert fr TNUMBER Tokens

istream* inp=&cin;           // aktueller Eingabedatenstrom
bool badtoken = false;       // true=Fehler gefunden
int curchar;                 // aktuelles Zeichen
TIncludeStack* istack = 0;   // offene Dateien, NULL = keine
bool freeform_flag = false;  // false = CR beendet Kommando, true = Freeform-Modus
string prompt;               // aktueller Prompt

bool at_eol = false;

bool noalias_flag = false;

extern volatile bool reading_stdin;

#ifdef WITH_READLINE
char* current_line = 0;  // aktuelle Zeile
char* current_char = 0;  // Zeichen darin
#endif

ostream& error_message()
{
    if(istack) {
	cerr << "Fehler in Zeile " << istack->lineno << ": ";
	closefile();
	at_eol=true;
    } else {
	cerr << "Fehler: ";
    }
    return cerr;
}

static char ungot_char = 0;

void ungetchar()
{
    ungot_char = curchar;
}

char getchar()
{
    if(ungot_char) {
        curchar = ungot_char;
        ungot_char = 0;
        return curchar;
    }
    
    static bool emit_prompt = true;
#ifdef WITH_READLINE
    static enum { RL_INIT, RL_OK, RL_EOL, RL_EOF } rl_status = RL_INIT;
	
    if(!istack) {
	if(rl_status==RL_INIT || rl_status==RL_EOL) {
	    if(current_line) free(current_line);
	    current_line = current_char = readline(prompt.c_str());
	    if(!current_line) rl_status = RL_EOF;
	    else {
		rl_status = RL_OK;
		if(*current_line) add_history(current_line);
	    }
	}
	if(rl_status==RL_EOF) {
	    if(current_line) free(current_line);
	    curchar=EOF;
	} else {
	    curchar = *current_char++;
	    if(curchar==0) {
		curchar = '\n';
		rl_status = RL_EOL;
	    }
	}
    }
    else curchar = inp->get();
#else
    if(!istack) {
	if(emit_prompt) cout << prompt;
	emit_prompt = false;
	reading_stdin = true;
    }
    curchar = inp->get();
#endif

    reading_stdin = false;
    at_eol = (curchar=='\n');
    if(at_eol) {
	if(!istack) emit_prompt = true;
	else istack->lineno++;
    }
    return curchar;
}

//
//  Liest ein Token und speichert es in ttype/tstr
//
TType getsymbol()
{
    if(badtoken && curchar!=EOF) return TBAD;
	
    while(1) {
	switch(curchar) {
         case '(':
	    getchar();
	    return ttype=TLEFT;
         case ')':
	    getchar();
	    return ttype=TRIGHT;
         case '?': case '\\':
	    getchar();
	    return ttype=TLAMBDA;
         case '.':
	    getchar();
	    return ttype=TDOT;
         case ':':
	    getchar();
	    return ttype=TCOLON;
         case '=':
	    if(noalias_flag || getchar()!='>') {
		badtoken=true;
		return ttype=TBAD;
	    }
	    getchar();
	    return ttype=TDOT;
         case '/':
	    getchar();
	    return ttype=TSLASH;
         case ';':
	    getchar();
	    return ttype=TSEMI;
         case '[':
	    getchar();
	    return ttype=TLEFTSQR;
         case ']':
	    getchar();
	    return ttype=TRIGHTSQR;
         case '0': case '1': case '2': case '3': case '4':
         case '5': case '6': case '7': case '8': case '9':
	    tint=0;
	    while(curchar!=EOF && isdigit(curchar)) {
		tint*=10;
		tint+=(curchar-'0');
		getchar();
	    }
	    return ttype=TNUMBER;
         case EOF: case 4: case 26: // damit man bei verkorksten Terminals
	    if(istack) {	   // mit C-d beenden kann, und auf DOSen mit C-z
		closefile();
		badtoken = false;
		break;
	    } else {
		badtoken = true;
		return ttype=TEOF;
	    }
         case '<':
	    tstr = "";
	    while(getchar()!='\n' && curchar!=EOF && curchar!='>')
		tstr += (char)curchar;
	    if(curchar=='\n') {
		error_message();
		cerr << "Zeilenende in Makroname." << endl;
		badtoken = true;
		return ttype=TBAD;
	    }
	    if(curchar==EOF) {
		error_message();
		cerr << "Eingabeende in Makroname." << endl;
		badtoken = true;
		return ttype=TBAD;
	    }
	    getchar();
	    return ttype=TMACRO;
         case '\n':
	    if(!freeform_flag) {
		curchar=' ';
		return ttype=TSEMI;
	    }
	    getchar();			    
	    break;
         case ' ': case '\t':
	    getchar();
	    break;
         case '#':
	    while(curchar!='\n' && curchar!=EOF) getchar();
	    if(curchar==EOF) {
		badtoken=true;
		return ttype=TEOF;
	    }
	    else getchar();
	    break;
         case '-':
	    if(getchar()!='>') {
		badtoken=true;
		return ttype=TBAD;
	    }
	    getchar();
	    return ttype=TARROW;
         case '\'': case '"':
            {
                int end=curchar;
                tstr="";
                while(getchar()!=EOF && curchar!='\n' && curchar!=end)
                    tstr += (char)curchar;
            }
            if(curchar==EOF) {
                cerr << "Fehler: Eingabeende in Name." << endl;
                badtoken=true;
                return ttype=TEOF;
            }
            if(curchar=='\n') {
                cerr << "Fehler: Zeilenende in Name." << endl;
                badtoken=true;
                return ttype=TBAD;
            }
            getchar();
            return ttype=TSYMBOL;
         default:
	    if(isalpha(curchar)) {
		tstr="";
		while(isalnum(curchar) || curchar=='-') {
		    tstr += (char)curchar;
		    getchar();
		}
                if(tstr[tstr.length()-1]=='-') {
                    /* HACK ALARM */
                    ungetchar();
                    curchar='-';
                    tstr.erase(tstr.length()-1);
                }
		if(!noalias_flag && tstr=="lambda") return ttype=TLAMBDA;
		return ttype=TSYMBOL;
	    }
	    error_message() << "Ungltiges Zeichen '" << (char)curchar
			    << "' (" << curchar << ")." << endl;
	    getchar();
	    badtoken = true;
	    return ttype=TBAD;
	}
    }
}

//
//  Startet den Scanner
//
void startscan()
{
    curchar=' ';
}

//
//  Scanner-Reset auf neuer Zeile
//
void resetscan()
{
    badtoken=false;
}

//
//  Zeile verwerfen nach Fehler
//
void readeol()
{
    if(!at_eol) while(curchar!='\n' && curchar!=EOF) getchar();
}

//
//  Datei ffnen (read)
//
void openstream(istream* s)
{
    if(!s) return;
    istack = new TIncludeStack(istack,s,curchar);
    inp = s;
    curchar = ' ';
}

void closefile()
{
    TIncludeStack* p = istack;
    istack = istack->prev;
    curchar = p->prevc;
    delete p->str;
    delete p;
    if(istack) inp=istack->str;
    else {
	inp=&cin;
    }
}

//
//  true, wenn die Eingabe == stdin ist
//
bool readtty()
{
    return istack==0;
}

//
//  true, wenn aktuelles Token das Symbol what ist
//
bool looking_at(const char* what)
{
    return ttype == TSYMBOL && tstr == what;
}
