//
//  Lambda: Symbol (freie/gebundene Variable)
//  . CSymbol
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//
#include "symbol.h"
#include "macro.h"
#include "application.h"

CNode* CSymbol::reduce()
{
    return copy();
}

CNode* CSymbol::eta(int stage, std::string s)
{
    if(stage == 1 && s == name)
        return this;
    else
        return 0;
}

CNode* CSymbol::copy()
{
    return new CSymbol(name);
}

void CSymbol::print(std::ostream& out, TPrintContext pc)
{
    out << name;
}

TVariable* CSymbol::freevars(TVariable* p)
{
    WHERE;
    TVariable* q = p;
    while(q) {
	if(q->name == name) return p;
	q=q->next;
    }
    return new TVariable(p,name);
}

bool CSymbol::alpha(TVariable* fv)
{
    while(fv) {
	if(fv->name == name && fv->replaceby!="") {
	    name = fv->replaceby;
	    return true;
	}
	fv = fv->next;
    }
    return false;
}

void CSymbol::global_alpha(TVariable* fv, int n)
{
    alpha(fv);
}

CNode* CSymbol::replace(std::string& sym, CNode* rep)
{
    if(name==sym) return rep->copy();
    return copy();
}

bool CSymbol::numcheck(TNumCheck& nc)
{
    if(nc.status==TNumCheck::CHECK_LEFT) {	// ?x.?y.x(x(x y))
	return name==nc.x;		        //       ^^
    } else if(nc.status==TNumCheck::CHECK_RIGHT) { // ?x.?y.x(x(x y))
	if(name==nc.x && nc.value==0) {            //             ^^
	    nc.value=-1;
	    return true;
	}
	return name==nc.y;
    } else if(nc.status==TNumCheck::CHECK_Y) {
        if(name==nc.x && nc.value==0) {
            nc.value=1;
            return true;
        }
        return false;
    } else
	return false;
}

bool CSymbol::listcheck(TListCheck& nc)
{
    if(nc.status==TListCheck::CHECK_1ST) {  // ?f.(f a tail)
	return name==nc.f;                  //     ^
    }
    else if(nc.status==TListCheck::CHECK_TAIL) {
	return name==nc.nil;
    }
    else if(nc.status==TListCheck::CHECK_2ND) {
	return true;
    }
    return false;
}

void CSymbol::listout(std::ostream& o, int n, char delim)
{
    if(n != TListCheck::CHECK_TAIL)
        print(o,PRINT_UNDEF);
}

CTypePtr CSymbol::get_type(TTypeInfo& ti, TTypeItem* local_types)
{
    for(TTypeItem* p = local_types; p != 0; p = p->next) {
        if(p->name == name)
            return p->type;
    }

    map<string, CTypePtr>::iterator i = ti.free_types.find(name);
    if(i != ti.free_types.end())
        return i->second;
    else {
        CTypePtr p = new CType();
        ti.free_types[name] = p;
        return p;
    }
}

CNode* CSymbol::ski()
{
    return new CSymbol(name);
}

CNode* CSymbol::ski_lambda(std::string s)
{
    if(s == name)
        return new CMacro("I");
    else
        return new CApplication(new CMacro("K"), new CSymbol(name));
}
